@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -20,6 +20,10 @@
 #ifndef gscolor1_INCLUDED
 #  define gscolor1_INCLUDED
 
+#include "stdpre.h"
+#include "gxtmap.h"
+#include "gsgstate.h"
+
 /* Color and gray interface */
 int gs_setcmykcolor(gs_gstate *, double, double, double, double),
     gs_currentcmykcolor(const gs_gstate *, float[4]),
