@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2020 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -36,6 +36,7 @@
 #include "string_.h"
 #include "strmio.h"         /* needed for sfclose */
 #include "gsicc_cache.h"    /* Needed for gsicc_get_icc_buff_hash */
+#include "gxdevsop.h"
 
 static cs_proc_install_cspace(gx_install_DeviceGray);
 static cs_proc_install_cspace(gx_install_DeviceRGB);
@@ -92,6 +93,8 @@ public_st_base_color_space();
 
 /* ------ Create/copy/destroy ------ */
 
+/* Ghostscript object finalizers can be called many times and hence
+ * must be idempotent. */
 static void
 gs_cspace_final(const gs_memory_t *cmem, void *vptr)
 {
@@ -100,17 +103,22 @@ gs_cspace_final(const gs_memory_t *cmem, void *vptr)
 
     if (pcs->interpreter_free_cspace_proc != NULL) {
         (*pcs->interpreter_free_cspace_proc) ((gs_memory_t *)cmem, pcs);
+        pcs->interpreter_free_cspace_proc = NULL;
     }
     if (pcs->type->final)
         pcs->type->final(pcs);
     if_debug2m('c', cmem, "[c]cspace final "PRI_INTPTR" %d\n", (intptr_t)pcs, (int)pcs->id);
     rc_decrement_only_cs(pcs->base_space, "gs_cspace_final");
-    if (pcs->params.device_n.devn_process_space != NULL)
-        rc_decrement_only_cs(pcs->params.device_n.devn_process_space, "gs_cspace_final");
+    pcs->base_space = NULL;
+    if (gs_color_space_get_index(pcs) == gs_color_space_index_DeviceN) {
+        if (pcs->params.device_n.devn_process_space != NULL) {
+            rc_decrement_only_cs(pcs->params.device_n.devn_process_space, "gs_cspace_final");
+            pcs->params.device_n.devn_process_space = NULL;
+        }
+    }
     /* No need to decrement the ICC profile data.  It is handled
        by the finalize of the ICC space which is called above using
        pcs->type->final(pcs);  */
-
 }
 
 static gs_color_space *
@@ -130,8 +138,10 @@ gs_cspace_alloc_with_id(gs_memory_t *mem, ulong id,
     pcs->interpreter_data = NULL;
     pcs->interpreter_free_cspace_proc = NULL;
     pcs->cmm_icc_profile_data = NULL;
+    pcs->ICC_Alternate_space = gs_ICC_Alternate_None;
     pcs->icc_equivalent = NULL;
     pcs->params.device_n.devn_process_space = NULL;
+    pcs->params.device_n.all_none = false;
     return pcs;
 }
 
@@ -425,7 +435,18 @@ void cs_adjust_counts_icc(gs_gstate *pgs, int delta)
     gs_color_space *pcs = gs_currentcolorspace_inline(pgs);
 
     if (pcs) {
-        cs_adjust_counts(pgs, delta);
+        cs_adjust_color_count(pgs, delta);
+        rc_adjust_const(pcs, delta, "cs_adjust_counts_icc");
+    }
+}
+
+void cs_adjust_swappedcounts_icc(gs_gstate *pgs, int delta)
+{
+    gs_color_space *pcs = gs_swappedcolorspace_inline(pgs);
+
+    if (pcs) {
+        cs_adjust_swappedcolor_count(pgs, delta);
+        rc_adjust_const(pcs, delta, "cs_adjust_swappedcounts_icc");
     }
 }
 
@@ -497,23 +518,14 @@ gx_set_no_overprint(gs_gstate* pgs)
 
 /* Retain all the spot colorants and not the process
    colorants.  This occurs if we have a process color
-   mismatch between the source and the destination but 
+   mismatch between the source and the destination but
    the output device supports spot colors */
 int
 gx_set_spot_only_overprint(gs_gstate* pgs)
 {
     gs_overprint_params_t   params = { 0 };
     gx_device* dev = pgs->device;
-    gx_color_index drawn_comps = 0;
-    gx_device_color_info* pcinfo = (dev == 0 ? 0 : &dev->color_info);
-
-    if (dev) {
-        /* check if color model behavior must be determined */
-        if (pcinfo->opmode == GX_CINFO_OPMODE_UNKNOWN)
-            drawn_comps = check_cmyk_color_model_comps(dev);
-        else
-            drawn_comps = pcinfo->process_comps;
-    }
+    gx_color_index drawn_comps = dev == NULL ? 0 : gx_get_process_comps(dev);
 
     params.retain_any_comps = true;
     params.op_state = OP_STATE_NONE;
@@ -586,10 +598,19 @@ check_cmyk_color_model_comps(gx_device * dev)
     gx_device_color_info *          pcinfo = &dev->color_info;
     uchar                           ncomps = pcinfo->num_components;
     int                             cyan_c, magenta_c, yellow_c, black_c;
-    subclass_color_mappings         scm;
     frac                            frac_14 = frac_1 / 4;
     frac                            out[GX_DEVICE_COLOR_MAX_COMPONENTS];
     gx_color_index                  process_comps;
+    const gx_cm_color_map_procs    *cmprocs;
+    const gx_device                *cmdev;
+
+
+    if (pcinfo->num_components < 4                     ||
+        pcinfo->polarity == GX_CINFO_POLARITY_ADDITIVE ||
+        pcinfo->gray_index == GX_CINFO_COMP_NO_INDEX) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
+        return 0;
+    }
 
     /* check for the appropriate components */
     if ( ncomps < 4                                       ||
@@ -597,54 +618,146 @@ check_cmyk_color_model_comps(gx_device * dev)
                        dev,
                        "Cyan",
                        sizeof("Cyan") - 1,
-                       NO_COMP_NAME_TYPE )) < 0           ||
+                       NO_COMP_NAME_TYPE_OP)) < 0           ||
          cyan_c == GX_DEVICE_COLOR_MAX_COMPONENTS         ||
          (magenta_c = dev_proc(dev, get_color_comp_index)(
                           dev,
                           "Magenta",
                           sizeof("Magenta") - 1,
-                          NO_COMP_NAME_TYPE )) < 0        ||
+                          NO_COMP_NAME_TYPE_OP)) < 0        ||
          magenta_c == GX_DEVICE_COLOR_MAX_COMPONENTS      ||
          (yellow_c = dev_proc(dev, get_color_comp_index)(
                         dev,
                         "Yellow",
                         sizeof("Yellow") - 1,
-                        NO_COMP_NAME_TYPE )) < 0               ||
+                        NO_COMP_NAME_TYPE_OP)) < 0               ||
          yellow_c == GX_DEVICE_COLOR_MAX_COMPONENTS       ||
          (black_c = dev_proc(dev, get_color_comp_index)(
                         dev,
                         "Black",
                         sizeof("Black") - 1,
-                        NO_COMP_NAME_TYPE )) < 0                         ||
+                        NO_COMP_NAME_TYPE_OP)) < 0                         ||
          black_c == GX_DEVICE_COLOR_MAX_COMPONENTS          )
         return 0;
 
     /* check the mapping */
-    scm = get_color_mapping_procs_subclass(dev);
+    cmprocs = dev_proc(dev, get_color_mapping_procs)(dev, &cmdev);
 
-    map_cmyk_subclass(scm, frac_14, frac_0, frac_0, frac_0, out);
-    if (!check_single_comp(cyan_c, frac_14, ncomps, out))
+    ncomps -= device_encodes_tags(dev);
+    cmprocs->map_cmyk(cmdev, frac_14, frac_0, frac_0, frac_0, out);
+    if (!check_single_comp(cyan_c, frac_14, ncomps, out)) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
         return 0;
-    map_cmyk_subclass(scm, frac_0, frac_14, frac_0, frac_0, out);
-    if (!check_single_comp(magenta_c, frac_14, ncomps, out))
+    }
+    cmprocs->map_cmyk(cmdev, frac_0, frac_14, frac_0, frac_0, out);
+    if (!check_single_comp(magenta_c, frac_14, ncomps, out)) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
         return 0;
-    map_cmyk_subclass(scm, frac_0, frac_0, frac_14, frac_0, out);
-    if (!check_single_comp(yellow_c, frac_14, ncomps, out))
-        return false;
-    map_cmyk_subclass(scm, frac_0, frac_0, frac_0, frac_14, out);
-    if (!check_single_comp(black_c, frac_14, ncomps, out))
+    }
+    cmprocs->map_cmyk(cmdev, frac_0, frac_0, frac_14, frac_0, out);
+    if (!check_single_comp(yellow_c, frac_14, ncomps, out)) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
         return 0;
+    }
+    cmprocs->map_cmyk(cmdev, frac_0, frac_0, frac_0, frac_14, out);
+    if (!check_single_comp(black_c, frac_14, ncomps, out)) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
+        return 0;
+    }
 
     process_comps =  ((gx_color_index)1 << cyan_c)
                    | ((gx_color_index)1 << magenta_c)
                    | ((gx_color_index)1 << yellow_c)
                    | ((gx_color_index)1 << black_c);
-    pcinfo->opmode = GX_CINFO_OPMODE;
+    pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED;
     pcinfo->process_comps = process_comps;
     pcinfo->black_component = black_c;
     return process_comps;
 }
 
+void
+check_rgb_color_model_comps(gx_device * dev)
+{
+    gx_device_color_info *          pcinfo = &dev->color_info;
+    uchar                           ncomps = pcinfo->num_components;
+    int                             red_c, green_c, blue_c;
+    frac                            frac_14 = frac_1 / 4;
+    frac                            out[GX_DEVICE_COLOR_MAX_COMPONENTS];
+    gx_color_index                  process_comps;
+    const gx_cm_color_map_procs    *cmprocs;
+    const gx_device                *cmdev;
+
+
+    if (pcinfo->num_components < 3                     ||
+        pcinfo->polarity != GX_CINFO_POLARITY_ADDITIVE ||
+        pcinfo->gray_index == GX_CINFO_COMP_NO_INDEX   ||
+        dev_proc(dev, dev_spec_op)(dev, gxdso_is_sep_supporting_additive_device, NULL, 0) <= 0
+        ) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
+        return;
+    }
+
+    /* check for the appropriate components */
+    if ( ncomps < 3                                      ||
+         (red_c = dev_proc(dev, get_color_comp_index)(
+                       dev,
+                       "Red",
+                       sizeof("Red") - 1,
+                       NO_COMP_NAME_TYPE_OP)) < 0        ||
+         red_c == GX_DEVICE_COLOR_MAX_COMPONENTS         ||
+         (green_c = dev_proc(dev, get_color_comp_index)(
+                          dev,
+                          "Green",
+                          sizeof("Green") - 1,
+                          NO_COMP_NAME_TYPE_OP)) < 0     ||
+         green_c == GX_DEVICE_COLOR_MAX_COMPONENTS       ||
+         (blue_c = dev_proc(dev, get_color_comp_index)(
+                        dev,
+                        "Blue",
+                        sizeof("Blue") - 1,
+                        NO_COMP_NAME_TYPE_OP)) < 0       ||
+         blue_c == GX_DEVICE_COLOR_MAX_COMPONENTS)
+    {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
+        return;
+    }
+
+    /* check the mapping */
+    cmprocs = dev_proc(dev, get_color_mapping_procs)(dev, &cmdev);
+
+    ncomps -= device_encodes_tags(dev);
+    cmprocs->map_rgb(cmdev, NULL, frac_14, frac_0, frac_0, out);
+    if (!check_single_comp(red_c, frac_14, ncomps, out)) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
+        return;
+    }
+    cmprocs->map_rgb(cmdev, NULL, frac_0, frac_14, frac_0, out);
+    if (!check_single_comp(green_c, frac_14, ncomps, out)) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
+        return;
+    }
+    cmprocs->map_rgb(cmdev, NULL, frac_0, frac_0, frac_14, out);
+    if (!check_single_comp(blue_c, frac_14, ncomps, out)) {
+        pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED_NOT;
+        return;
+    }
+
+    process_comps =  ((gx_color_index)1 << red_c)
+                   | ((gx_color_index)1 << green_c)
+                   | ((gx_color_index)1 << blue_c);
+    pcinfo->opmsupported = GX_CINFO_OPMSUPPORTED;
+    pcinfo->process_comps = process_comps;
+    pcinfo->black_component = 0; /* ? */
+}
+
+void check_opmsupported(gx_device * dev)
+{
+    if (dev->color_info.polarity == GX_CINFO_POLARITY_ADDITIVE)
+        check_rgb_color_model_comps(dev);
+    else
+        (void)check_cmyk_color_model_comps(dev);
+}
+
 /*
  * This set_overprint method is unique. If overprint is true, overprint
  * mode is set to 1, the process color model has DeviceCMYK behavior (see
@@ -662,7 +775,7 @@ gx_set_overprint_DeviceCMYK(const gs_color_space * pcs, gs_gstate * pgs)
     if ( !pgs->overprint                      ||
          pgs->overprint_mode != 1             ||
          pcinfo == 0                          ||
-         pcinfo->opmode == GX_CINFO_OPMODE_NOT  )
+         pcinfo->opmsupported == GX_CINFO_OPMSUPPORTED_NOT)
         return gx_spot_colors_set_overprint(pcs, pgs);
     /* Share code with CMYK ICC case */
     return gx_set_overprint_cmyk(pcs, pgs);
@@ -686,7 +799,6 @@ gx_set_overprint_DeviceCMYK(const gs_color_space * pcs, gs_gstate * pgs)
 int gx_set_overprint_cmyk(const gs_color_space * pcs, gs_gstate * pgs)
 {
     gx_device *             dev = pgs->device;
-    gx_device_color_info *  pcinfo = (dev == 0 ? 0 : &dev->color_info);
     gx_color_index          drawn_comps = 0;
     gs_overprint_params_t   params = { 0 };
     gx_device_color        *pdc;
@@ -708,11 +820,7 @@ int gx_set_overprint_cmyk(const gs_color_space * pcs, gs_gstate * pgs)
         gsicc_extract_profile(dev->graphics_type_tag, dev_profile, &(output_profile),
                               &render_cond);
 
-        /* check if color model behavior must be determined */
-        if (pcinfo->opmode == GX_CINFO_OPMODE_UNKNOWN)
-            drawn_comps = check_cmyk_color_model_comps(dev);
-        else
-            drawn_comps = pcinfo->process_comps;
+        drawn_comps = gx_get_process_comps(dev);
     }
 
     if_debug1m(gs_debug_flag_overprint, pgs->memory,
@@ -724,8 +832,7 @@ int gx_set_overprint_cmyk(const gs_color_space * pcs, gs_gstate * pgs)
     /* correct for any zero'ed color components.  But only if profiles
        match AND pgs->overprint_mode is true */
     if (pcs->cmm_icc_profile_data != NULL && output_profile != NULL) {
-        if (output_profile->hashcode ==
-            pcs->cmm_icc_profile_data->hashcode) {
+        if (gsicc_profiles_equal(output_profile, pcs->cmm_icc_profile_data)) {
             profile_ok = true;
         }
     }
@@ -759,13 +866,13 @@ int gx_set_overprint_cmyk(const gs_color_space * pcs, gs_gstate * pgs)
                space has to be CMYK. Trick is that we do need to worry about
                the colorant order on the target device */
             num_colorant[0] = (dev_proc(dev, get_color_comp_index))\
-                             (dev, "Cyan", strlen("Cyan"), NO_COMP_NAME_TYPE);
+                             (dev, "Cyan", strlen("Cyan"), NO_COMP_NAME_TYPE_OP);
             num_colorant[1] = (dev_proc(dev, get_color_comp_index))\
-                             (dev, "Magenta", strlen("Magenta"), NO_COMP_NAME_TYPE);
+                             (dev, "Magenta", strlen("Magenta"), NO_COMP_NAME_TYPE_OP);
             num_colorant[2] = (dev_proc(dev, get_color_comp_index))\
-                             (dev, "Yellow", strlen("Yellow"), NO_COMP_NAME_TYPE);
+                             (dev, "Yellow", strlen("Yellow"), NO_COMP_NAME_TYPE_OP);
             num_colorant[3] = (dev_proc(dev, get_color_comp_index))\
-                             (dev, "Black", strlen("Black"), NO_COMP_NAME_TYPE);
+                             (dev, "Black", strlen("Black"), NO_COMP_NAME_TYPE_OP);
             nz_comps = 0;
             one = 1;
             colorant_ok = true;
@@ -809,6 +916,123 @@ int gx_set_overprint_cmyk(const gs_color_space * pcs, gs_gstate * pgs)
     return gs_gstate_update_overprint(pgs, &params);
 }
 
+int gx_set_overprint_rgb(const gs_color_space * pcs, gs_gstate * pgs)
+{
+    gx_device *             dev = pgs->device;
+    gx_color_index          drawn_comps = 0;
+    gs_overprint_params_t   params = { 0 };
+    gx_device_color        *pdc;
+    cmm_dev_profile_t      *dev_profile;
+    cmm_profile_t          *output_profile = 0;
+    int                     code;
+    bool                    profile_ok = false;
+    gsicc_rendering_param_t        render_cond;
+    bool                    eop;
+
+    if_debug0m(gs_debug_flag_overprint, pgs->memory,
+        "[overprint] gx_set_overprint_rgb\n");
+
+    if (dev) {
+        code = dev_proc(dev, get_profile)(dev, &dev_profile);
+        if (code < 0)
+            return code;
+
+        gsicc_extract_profile(dev->graphics_type_tag, dev_profile, &(output_profile),
+                              &render_cond);
+
+        drawn_comps = gx_get_process_comps(dev);
+    }
+
+    if_debug1m(gs_debug_flag_overprint, pgs->memory,
+        "[overprint] gx_set_overprint_rgb. drawn_comps = 0x%x\n", (uint)drawn_comps);
+
+    if (drawn_comps == 0)
+        return gx_spot_colors_set_overprint(pcs, pgs);
+
+    /* correct for any zero'ed color components.  But only if profiles
+       match AND pgs->overprint_mode is true */
+    if (pcs->cmm_icc_profile_data != NULL && output_profile != NULL) {
+        if (gsicc_profiles_equal(output_profile, pcs->cmm_icc_profile_data)) {
+            profile_ok = true;
+        }
+    }
+
+    eop = gs_currentcolor_eopm(pgs);
+
+    if_debug3m(gs_debug_flag_overprint, pgs->memory,
+        "[overprint] gx_set_overprint_rgb. is_fill_color = %d, pgs->color[0].effective_opm = %d pgs->color[1].effective_opm = %d\n",
+        pgs->is_fill_color, pgs->color[0].effective_opm, pgs->color[1].effective_opm);
+
+    if (profile_ok && eop) {
+        gx_color_index  nz_comps, one, temp;
+        int             code;
+        int             num_colorant[3], k;
+        bool            colorant_ok;
+        dev_color_proc_get_nonzero_comps((*procp));
+
+        if_debug0m(gs_debug_flag_overprint, pgs->memory,
+            "[overprint] gx_set_overprint_cmyk. color_is_set, profile_ok and eop\n");
+
+        code = gx_set_dev_color(pgs);
+        if (code < 0)
+            return code;
+        pdc = gs_currentdevicecolor_inline(pgs);
+        procp = pdc->type->get_nonzero_comps;
+        if (pdc->ccolor_valid) {
+            /* If we have the source colors, then use those in making the
+               decision as to which ones are non-zero.  Then we avoid
+               accidently looking at small values that get quantized to zero
+               Note that to get here in the code, the source color data color
+               space has to be CMYK. Trick is that we do need to worry about
+               the colorant order on the target device */
+            num_colorant[0] = (dev_proc(dev, get_color_comp_index))\
+                             (dev, "Red", strlen("Red"), NO_COMP_NAME_TYPE_OP);
+            num_colorant[1] = (dev_proc(dev, get_color_comp_index))\
+                             (dev, "Green", strlen("Green"), NO_COMP_NAME_TYPE_OP);
+            num_colorant[2] = (dev_proc(dev, get_color_comp_index))\
+                             (dev, "Blue", strlen("Blue"), NO_COMP_NAME_TYPE_OP);
+            nz_comps = 0;
+            one = 1;
+            colorant_ok = true;
+            for (k = 0; k < 3; k++) {
+                /* Note: AR assumes the value is zero if it
+                   is less than 0.5 out of 255 */
+                if (pdc->ccolor.paint.values[k] > (0.5 / 255.0)) {
+                    if (num_colorant[k] == -1) {
+                        colorant_ok = false;
+                    } else {
+                        temp = one << num_colorant[k];
+                        nz_comps = nz_comps | temp;
+                    }
+                }
+            }
+            /* For some reason we don't have one of the standard colorants */
+            if (!colorant_ok) {
+                if ((code = procp(pdc, dev, &nz_comps)) < 0)
+                    return code;
+            }
+        } else {
+            if ((code = procp(pdc, dev, &nz_comps)) < 0)
+                return code;
+        }
+        drawn_comps &= nz_comps;
+    }
+    params.is_fill_color = pgs->is_fill_color;
+    params.retain_any_comps = true;
+    params.drawn_comps = drawn_comps;
+    params.op_state = OP_STATE_NONE;
+
+    if_debug2m(gs_debug_flag_overprint, pgs->memory,
+        "[overprint] gx_set_overprint_rgb. retain_any_comps = %d, drawn_comps = 0x%x\n",
+        params.retain_any_comps, (uint)(params.drawn_comps));
+
+    /* We are in RGB, the profiles match and overprint is true.  Set effective
+       overprint mode to overprint mode but only if effective has not already
+       been set to 0 */
+    params.effective_opm = pgs->color[0].effective_opm =
+        pgs->overprint_mode && gs_currentcolor_eopm(pgs);
+    return gs_gstate_update_overprint(pgs, &params);
+}
 /* A stub for a color mapping linearity check, when it is inapplicable. */
 int
 gx_cspace_no_linear(const gs_color_space *cs, const gs_gstate * pgs,
@@ -1027,8 +1251,13 @@ ENUM_PTRS_BEGIN_PROC(color_space_enum_ptrs)
         return ENUM_OBJ(pcs->pclient_color_space_data);
     if (index == 2)
         return ENUM_OBJ(pcs->icc_equivalent);
-    if (index == 3)
-        return ENUM_OBJ(pcs->params.device_n.devn_process_space);
+    if (index == 3) {
+        if (gs_color_space_get_index(pcs) == gs_color_space_index_DeviceN)
+            return ENUM_OBJ(pcs->params.device_n.devn_process_space);
+        else
+            return ENUM_OBJ(NULL);
+    }
+
     return ENUM_USING(*pcs->type->stype, vptr, size, index - 4);
     ENUM_PTRS_END_PROC
 }
@@ -1038,7 +1267,8 @@ RELOC_PTRS_WITH(color_space_reloc_ptrs, gs_color_space *pcs)
     RELOC_VAR(pcs->base_space);
     RELOC_VAR(pcs->pclient_color_space_data);
     RELOC_VAR(pcs->icc_equivalent);
-    RELOC_VAR(pcs->params.device_n.devn_process_space);
+    if (gs_color_space_get_index(pcs) == gs_color_space_index_DeviceN)
+        RELOC_VAR(pcs->params.device_n.devn_process_space);
     RELOC_USING(*pcs->type->stype, vptr, size);
 }
 RELOC_PTRS_END
