@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2023 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -42,12 +42,12 @@ extern const gx_device_color_type_t gx_dc_pattern2;
  */
 #define ENCODE_VALUE(v, emax, vmin, vmax)\
   ( ((v) - (vmin)) * ((double)(emax) / ((vmax) - (vmin))) )
-/*
- * Because of the Acrobat Reader limitation noted in gdevpdfx.h,
- * we must limit coordinate values to 14 bits.
- */
-#define MIN_MESH_COORDINATE (-0x400000 / 256.0)
-#define MAX_MESH_COORDINATE ( 0x3fffff / 256.0)
+#define PDFA_MIN_MESH_COORDINATE (-0x400000 / 128.0)
+#define PDFA_MAX_MESH_COORDINATE ( 0x3fffff / 128.0)
+#define PDFA_ENCODE_MESH_COORDINATE(v)\
+  ENCODE_VALUE(v, 0xffffff, PDFA_MIN_MESH_COORDINATE, PDFA_MAX_MESH_COORDINATE)
+#define MIN_MESH_COORDINATE (-0x800000 )
+#define MAX_MESH_COORDINATE ( 0x7fffff )
 #define ENCODE_MESH_COORDINATE(v)\
   ENCODE_VALUE(v, 0xffffff, MIN_MESH_COORDINATE, MAX_MESH_COORDINATE)
 
@@ -689,6 +689,8 @@ typedef struct pdf_mesh_data_params_s {
     int num_points;
     int num_components;
     bool is_indexed;
+    int rescale;            /* If the co-ordinates won't fit into crappy Acrobat values, scale them here and in the pattern Matrix */
+    bool old_pdf;
     const float *Domain;	/* iff Function */
     const gs_range_t *ranges;
 } pdf_mesh_data_params_t;
@@ -710,9 +712,13 @@ put_clamped(byte *p, double v, int num_bytes)
         *p++ = (byte)(i >> shift);
 }
 static inline void
-put_clamped_coord(byte *p, double v, int num_bytes)
+put_clamped_coord(byte *p, double v, int num_bytes, const pdf_mesh_data_params_t *pmdp)
 {
-    put_clamped(p, ENCODE_MESH_COORDINATE(v), num_bytes);
+    if (pmdp->rescale != 1.0 || pmdp->old_pdf) {
+        v = v / pmdp->rescale;
+        put_clamped(p, PDFA_ENCODE_MESH_COORDINATE(v), num_bytes);
+    } else
+        put_clamped(p, ENCODE_MESH_COORDINATE(v), num_bytes);
 }
 
 /* Convert floating-point mesh data to packed binary. */
@@ -733,8 +739,8 @@ put_float_mesh_data(gx_device_pdf *pdev, cos_stream_t *pscs, shade_coord_stream_
     if ((code = shade_next_coords(cs, pts, num_points)) < 0)
         return code;
     for (i = 0; i < num_points; ++i) {
-        put_clamped_coord(b + 1 + i * 6, fixed2float(pts[i].x), 3);
-        put_clamped_coord(b + 4 + i * 6, fixed2float(pts[i].y), 3);
+        put_clamped_coord(b + 1 + i * 6, fixed2float(pts[i].x), 3, pmdp);
+        put_clamped_coord(b + 4 + i * 6, fixed2float(pts[i].y), 3, pmdp);
     }
     if ((code = cos_stream_add_bytes(pdev, pscs, b + (flag < 0),
                                      (flag >= 0) + num_points * 6)) < 0)
@@ -780,7 +786,7 @@ put_float_mesh_data(gx_device_pdf *pdev, cos_stream_t *pscs, shade_coord_stream_
 /* Write a mesh Shading. */
 static int
 pdf_put_mesh_shading(gx_device_pdf *pdev, cos_stream_t *pscs, const gs_shading_t *psh,
-                     const gs_range_t *pranges)
+                     const gs_range_t *pranges, int *rescale)
 {
     cos_dict_t *const pscd = cos_stream_dict(pscs);
     gs_color_space *pcs = psh->params.ColorSpace;
@@ -803,6 +809,8 @@ pdf_put_mesh_shading(gx_device_pdf *pdev, cos_stream_t *pscs, const gs_shading_t
         num_comp = gs_color_space_num_components(pcs);
     }
     data_params.ranges = pranges;
+    data_params.rescale = 1;
+
 
     /* Write parameters common to all mesh Shadings. */
     shade_next_init(&cs, pmp, NULL);
@@ -812,10 +820,18 @@ pdf_put_mesh_shading(gx_device_pdf *pdev, cos_stream_t *pscs, const gs_shading_t
 
         if (pca == 0)
             return_error(gs_error_VMerror);
-        for (i = 0; i < 2; ++i)
-            if ((code = pdf_array_add_real2(pca, MIN_MESH_COORDINATE,
+        for (i = 0; i < 2; ++i) {
+            if (pdev->CompatibilityLevel < 1.5) {
+                if ((code = pdf_array_add_real2(pca, PDFA_MIN_MESH_COORDINATE,
+                                            PDFA_MAX_MESH_COORDINATE)) < 0)
+                    return code;
+            }
+            else {
+                if ((code = pdf_array_add_real2(pca, MIN_MESH_COORDINATE,
                                             MAX_MESH_COORDINATE)) < 0)
-                return code;
+                    return code;
+            }
+        }
         data_params.is_indexed = false;
         if (gs_color_space_get_index(pcs) == gs_color_space_index_Indexed) {
             data_params.is_indexed = true;
@@ -872,6 +888,150 @@ pdf_put_mesh_shading(gx_device_pdf *pdev, cos_stream_t *pscs, const gs_shading_t
         )
         return code;
 
+    if (from_array && pdev->CompatibilityLevel < 1.5 ) {
+        float min_x, max_x, min_y, max_y, z;
+        int i = 0, j, num_points = 1, num_components = 1;
+        float c = 0;
+        float x, y;
+
+        /*
+         * Because of the Acrobat Reader limitation noted in gdevpdfx.h,
+         * we must limit coordinate values to 16 bits. We no longer
+         * attempt to support Acrobat 1.3 and below (which only permit 14 bits).
+         */
+        min_x = min_y = PDFA_MIN_MESH_COORDINATE;
+        max_x = max_y = PDFA_MAX_MESH_COORDINATE;
+
+        switch(ShadingType(psh)){
+            case shading_type_Tensor_product_patch:
+                    while ((flag = shade_next_flag(&cs, 0)) >= 0){
+                        num_points = (flag == 0 ? 16 : 12);
+                        num_components = num_comp * (flag == 0 ? 4 : 2);
+
+                        for (j = 0; j < num_points;j++) {
+                            code = cs.get_decoded(&cs, 0, NULL, &x);
+                            if (code < 0)
+                                break;
+                            code = cs.get_decoded(&cs, 0, NULL, &y);
+                            if (code < 0)
+                                break;
+                            for (i = 0; i < num_components; ++i) {
+                                code = cs.get_decoded(&cs, 0, NULL, &c);
+                                if (code < 0)
+                                    break;
+                            }
+                            if (x < min_x)
+                                min_x = x;
+                            if (y < min_y)
+                                min_y = y;
+                            if (x > max_x)
+                                max_x = x;
+                            if (y > max_y)
+                                max_y = y;
+                        }
+                    }
+                break;
+            case shading_type_Coons_patch:
+                    while ((flag = shade_next_flag(&cs, 0)) >= 0){
+                        num_points = (flag == 0 ? 12 : 8);
+                        num_components = num_comp * (flag == 0 ? 4 : 2);
+
+                        for (j = 0; j < num_points;j++) {
+                            code = cs.get_decoded(&cs, 0, NULL, &x);
+                            if (code < 0)
+                                break;
+                            code = cs.get_decoded(&cs, 0, NULL, &y);
+                            if (code < 0)
+                                break;
+                            for (i = 0; i < num_components; ++i) {
+                                code = cs.get_decoded(&cs, 0, NULL, &c);
+                                if (code < 0)
+                                    break;
+                            }
+                            if (x < min_x)
+                                min_x = x;
+                            if (y < min_y)
+                                min_y = y;
+                            if (x > max_x)
+                                max_x = x;
+                            if (y > max_y)
+                                max_y = y;
+                        }
+                    }
+                break;
+            case shading_type_Free_form_Gouraud_triangle:
+                {
+                    while ((flag = shade_next_flag(&cs, 0)) >= 0){
+                        code = cs.get_decoded(&cs, 0, NULL, &x);
+                        if (code < 0)
+                            break;
+                        code = cs.get_decoded(&cs, 0, NULL, &y);
+                        if (code < 0)
+                            break;
+                        for (i = 0; i < num_comp; ++i) {
+                            code = cs.get_decoded(&cs, 0, NULL, &c);
+                            if (code < 0)
+                                break;
+                        }
+                        if (x < min_x)
+                            min_x = x;
+                        if (y < min_y)
+                            min_y = y;
+                        if (x > max_x)
+                            max_x = x;
+                        if (y > max_y)
+                            max_y = y;
+                    }
+                }
+                break;
+            case shading_type_Lattice_form_Gouraud_triangle:
+                {
+                    while (!seofp(cs.s)) {
+                        code = cs.get_decoded(&cs, 0, NULL, &x);
+                        if (code < 0)
+                            break;
+                        code = cs.get_decoded(&cs, 0, NULL, &y);
+                        if (code < 0)
+                            break;
+                        for (i = 0; i < num_comp; ++i) {
+                            code = cs.get_decoded(&cs, 0, NULL, &c);
+                            if (code < 0)
+                                break;
+                        }
+                        if (x < min_x)
+                            min_x = x;
+                        if (y < min_y)
+                            min_y = y;
+                        if (x > max_x)
+                            max_x = x;
+                        if (y > max_y)
+                            max_y = y;
+                    }
+                }
+                break;
+            default:
+                break;
+        }
+        s_init(&cs.ds, NULL);
+        sread_string(&cs.ds, pmp->DataSource.data.str.data,
+                     pmp->DataSource.data.str.size);
+        cs.s = &cs.ds;
+
+        *rescale = (int)ceil(min_x / PDFA_MIN_MESH_COORDINATE);
+        z = ceil(min_y / PDFA_MIN_MESH_COORDINATE);
+        if (z > *rescale)
+            *rescale = (int)z;
+        z = ceil(max_x / PDFA_MAX_MESH_COORDINATE);
+        if (z > *rescale)
+            *rescale = (int)z;
+        z = ceil(max_y / PDFA_MAX_MESH_COORDINATE);
+        if (z > *rescale)
+            *rescale = (int)z;
+        data_params.rescale = *rescale;
+        data_params.old_pdf = 1;
+    } else
+        data_params.old_pdf = 0;
+
     switch (ShadingType(psh)) {
     case shading_type_Free_form_Gouraud_triangle: {
         const gs_shading_FfGt_params_t *const params =
@@ -949,7 +1109,7 @@ pdf_put_mesh_shading(gx_device_pdf *pdev, cos_stream_t *pscs, const gs_shading_t
     }
     code1 =  cos_dict_put_c_key_int(pscd, "/BitsPerFlag", bits_per_flag);
     if (code1 < 0)
-        return code;
+        return code1;
     return code;
 }
 
@@ -969,7 +1129,7 @@ pdf_put_pattern2(gx_device_pdf *pdev, const gs_gstate * pgs, const gx_drawing_co
     cos_object_t *psco;
     const gs_range_t *pranges;
     int code = pdf_cs_Pattern_colored(pdev, &v);
-    int code1 = 0;
+    int code1 = 0, rescale = 1;
     gs_matrix smat;
     gs_point dist;
 
@@ -991,7 +1151,7 @@ pdf_put_pattern2(gx_device_pdf *pdev, const gs_gstate * pgs, const gx_drawing_co
         code = pdf_put_shading_common(pdev, cos_stream_dict((cos_stream_t *)psco), pgs,
                                       psh, pinst->shfill, &pranges);
         if (code >= 0)
-            code1 = pdf_put_mesh_shading(pdev, (cos_stream_t *)psco, psh, pranges);
+            code1 = pdf_put_mesh_shading(pdev, (cos_stream_t *)psco, psh, pranges, &rescale);
         else
             /* We won't use this shading, we fall back because we couldn't write it */
             psres->where_used = 0;
@@ -1028,6 +1188,11 @@ pdf_put_pattern2(gx_device_pdf *pdev, const gs_gstate * pgs, const gx_drawing_co
 
         smat.xx *= xscale, smat.yx *= xscale, smat.tx *= xscale;
         smat.xy *= yscale, smat.yy *= yscale, smat.ty *= yscale;
+
+        if (rescale != 1) {
+            smat.xx *= rescale, smat.yx *= rescale;
+            smat.xy *= rescale, smat.yy *= rescale;
+        }
     }
 
     /* Bug #697451, if we emit a PDF with a type 2 Pattern where the
