@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -29,6 +29,7 @@
 #include "gdevpsdf.h"
 #include "gxdevmem.h"
 #include "sarc4.h"
+#include "gsfunc.h"
 
 #define FINE_GLYPH_USAGE 1 /* Old code = 0, new code = 1 */
 
@@ -40,7 +41,17 @@
  * Windows and Linux fail with coordinates outside +/- 16383. Hence, we
  * limit coordinates to 16k, with a little slop.
  */
-#define MAX_USER_COORD 16300
+/* 28/05/2020 This was only being applied to text and a subset of paths. Since
+ * Acrobat 4 is now more than 20 years old, lets just drop support for it. The
+ * PDF specification never had this limit, just Adobe's software.
+ */
+/* 31/05/2021 However, it transpires that the PDF/A-1 specification adopted the
+ * implementation limits of Acrobat 5 as part of the specification. So we need
+ * to handle that, but we don't need to limit ourselves to the more restrictive
+ * Acrobat 4 limit of 16384 ( MAX_USER_COORD 16300), we can use 32767. THis is only
+ * used when creating PDF/A-1 files.
+ */
+#define MAX_USER_COORD 32700
 
 /* ---------------- Statically allocated sizes ---------------- */
 /* These should really be dynamic.... */
@@ -52,8 +63,13 @@
 /* Define the maximum size of a destination array string. */
 #define MAX_DEST_STRING 80
 
+/* Define the maximum number of objects stored in an ObjStm */
+#define MAX_OBJSTM_OBJECTS 200
+
 /* ================ Types and structures ================ */
 
+typedef struct pdf_base_font_s pdf_base_font_t;
+
 /* Define the possible contexts for the output stream. */
 typedef enum {
     PDF_IN_NONE,
@@ -77,15 +93,9 @@ typedef struct cos_object_procs_s cos_object_procs_t;
 typedef const cos_object_procs_t *cos_type_t;
 #define cos_types_DEFINED
 
-#ifndef pdf_text_state_DEFINED
-#  define pdf_text_state_DEFINED
 typedef struct pdf_text_state_s pdf_text_state_t;
-#endif
 
-#ifndef pdf_char_glyph_pairs_DEFINED
-#  define pdf_char_glyph_pairs_DEFINED
 typedef struct pdf_char_glyph_pairs_s pdf_char_glyph_pairs_t;
-#endif
 
 /* ---------------- Resources ---------------- */
 
@@ -186,8 +196,8 @@ typedef enum {
     gs_id rid;                        /* optional ID key */\
     bool named;\
     bool global;                /* ps2write only */\
-    char rname[1/*R*/ + (sizeof(long) * 8 / 3 + 1) + 1/*\0*/];\
-    ulong where_used;                /* 1 bit per level of content stream */\
+    char rname[1/*R*/ + (sizeof(int64_t) * 8 / 3 + 1) + 1/*\0*/];\
+    uint64_t where_used;                /* 1 bit per level of content stream */\
     cos_object_t *object
 typedef struct pdf_resource_s pdf_resource_t;
 struct pdf_resource_s {
@@ -236,7 +246,7 @@ typedef struct pdf_text_data_s pdf_text_data_t;  /* gdevpdft.h */
 
 /* Outline nodes and levels */
 typedef struct pdf_outline_node_s {
-    long id, parent_id, prev_id, first_id, last_id;
+    int64_t id, parent_id, prev_id, first_id, last_id;
     int count;
     cos_dict_t *action;
 } pdf_outline_node_t;
@@ -252,7 +262,7 @@ typedef struct pdf_outline_level_s {
 
 /* Articles */
 typedef struct pdf_bead_s {
-    long id, article_id, prev_id, next_id, page_id;
+    int64_t id, article_id, prev_id, next_id, page_id;
     gs_rect rect;
 } pdf_bead_t;
 typedef struct pdf_article_s pdf_article_t;
@@ -288,7 +298,7 @@ typedef struct pdf_resource_list_s {
 
 /* Define the bookkeeping for an open stream. */
 typedef struct pdf_stream_position_s {
-    long length_id;
+    int64_t length_id;
     gs_offset_t start_pos;
 } pdf_stream_position_t;
 
@@ -298,7 +308,7 @@ typedef struct pdf_stream_position_s {
  * and one for the whole document (for AutoRotate /All).
  */
 typedef struct pdf_text_rotation_s {
-    long counts[5];                /* 0, 90, 180, 270, other */
+    int64_t counts[5];                /* 0, 90, 180, 270, other */
     int Rotate;                        /* computed rotation, -1 means none */
 } pdf_text_rotation_t;
 #define pdf_text_rotation_angle_values 0, 90, 180, 270, -1
@@ -321,14 +331,15 @@ typedef struct pdf_page_s {
     cos_dict_t *Page;
     gs_point MediaBox;
     pdf_procset_t procsets;
-    long contents_id;
-    long resource_ids[resourceFont + 1]; /* resources thru Font, see above */
-    long group_id;
+    int64_t contents_id;
+    int64_t resource_ids[resourceFont + 1]; /* resources thru Font, see above */
+    int64_t group_id;
     cos_array_t *Annots;
     pdf_text_rotation_t text_rotation;
     pdf_page_dsc_info_t dsc_info;
     bool NumCopies_set; /* ps2write only. */
     int NumCopies;      /* ps2write only. */
+    float UserUnit;     /* pdfwrite only */
 } pdf_page_t;
 #define private_st_pdf_page()        /* in gdevpdf.c */\
   gs_private_st_ptrs2(st_pdf_page, pdf_page_t, "pdf_page_t",\
@@ -340,16 +351,13 @@ typedef struct pdf_page_s {
  */
 typedef struct pdf_temp_file_s {
     char file_name[gp_file_name_sizeof];
-    FILE *file;
+    gp_file *file;
     stream *strm;
     byte *strm_buf;
     stream *save_strm;                /* save pdev->strm while writing here */
 } pdf_temp_file_t;
 
-#ifndef gx_device_pdf_DEFINED
-#  define gx_device_pdf_DEFINED
 typedef struct gx_device_pdf_s gx_device_pdf;
-#endif
 
 /* Structures and definitions for linearisation */
 typedef struct linearisation_record_s {
@@ -417,22 +425,22 @@ typedef struct share_hint_stream_s {
 } shared_hint_stream_t;
 
 typedef struct pdf_linearisation_s {
-    FILE *sfile;
+    gp_file *sfile;
     pdf_temp_file_t Lin_File;
     char HintBuffer[256];
     unsigned char HintBits;
     unsigned char HintByte;
-    long Catalog_id;
-    long Info_id;
-    long Pages_id;
-    long NumPage1Resources;
-    long NumPart1StructureResources;
-    long NumSharedResources;
-    long NumUniquePageResources;
-    long NumPart9Resources;
-    long NumNonPageResources;
-    long LastResource;
-    long MainFileEnd;
+    int64_t Catalog_id;
+    int64_t Info_id;
+    int64_t Pages_id;
+    int64_t NumPage1Resources;
+    int64_t NumPart1StructureResources;
+    int64_t NumSharedResources;
+    int64_t NumUniquePageResources;
+    int64_t NumPart9Resources;
+    int64_t NumNonPageResources;
+    int64_t LastResource;
+    int64_t MainFileEnd;
     gs_offset_t *Offsets;
     gs_offset_t xref;
     gs_offset_t FirstxrefOffset;
@@ -485,8 +493,9 @@ struct pdf_font_cache_elem_s {
 typedef struct pdf_viewer_state_s {
     int transfer_not_identity;        /* bitmask */
     gs_id transfer_ids[4];
-    float opacity_alpha; /* state.opacity.alpha */
-    float shape_alpha; /* state.shape.alpha */
+    float strokeconstantalpha;
+    float fillconstantalpha;
+    bool alphaisshape;
     gs_blend_mode_t blend_mode; /* state.blend_mode */
     gs_id halftone_id;
     gs_id black_generation_id;
@@ -497,7 +506,7 @@ typedef struct pdf_viewer_state_s {
     bool text_knockout; /* state.text_knockout */
     bool fill_overprint;
     bool stroke_overprint;
-    bool stroke_adjust; /* state.stroke_adjust */
+    int stroke_adjust; /* state.stroke_adjust */
     bool fill_used_process_color;
     bool stroke_used_process_color;
     gx_hl_saved_color saved_fill_color;
@@ -547,6 +556,33 @@ typedef enum {
     pdf_compress_Flate
 } pdf_compression_type;
 
+typedef enum {
+    OCR_UnInit,
+    OCR_Rendering,
+    OCR_Rendered,
+    OCR_UnicodeAvailable,
+    OCR_Failed
+} pdf_OCR_stage;
+
+typedef enum {
+    UseOCRNever,
+    UseOCRAsNeeded,
+    UseOCRAlways
+} pdf_OCR_usage;
+
+typedef struct ocr_glyph_s{
+    byte *data;
+    int x;
+    int y;
+    int width;
+    int height;
+    int raster;
+    void *next;
+    gs_char char_code;
+    gs_glyph glyph;
+    bool is_space;
+} ocr_glyph_t;
+
 /* Define the device structure. */
 struct gx_device_pdf_s {
     gx_device_psdf_common;
@@ -572,10 +608,13 @@ struct gx_device_pdf_s {
     gs_param_float_array PDFXTrimBoxToMediaBoxOffset;
     gs_param_float_array PDFXBleedBoxToTrimBoxOffset;
     bool PDFXSetBleedBoxToMediaBox;
+    /* OCR Parameters */
+    char ocr_language[1024];
+    int ocr_engine;
     /* Other parameters */
     bool ReAssignCharacters;
     bool ReEncodeCharacters;
-    long FirstObjectNumber;
+    int64_t FirstObjectNumber;
     bool CompressFonts;
     bool CompressStreams;
     bool PrintStatistics;
@@ -595,17 +634,17 @@ struct gx_device_pdf_s {
     bool PDFX;                   /* Generate PDF/X */
     int PDFA;                   /* Generate PDF/A 0 = don't produce, otherwise level of PDF/A */
     bool AbortPDFAX;            /* Abort generation of PDFA or X, produce regular PDF */
-    long MaxClipPathSize;  /* The maximal number of elements of a clipping path
+    int64_t MaxClipPathSize;  /* The maximal number of elements of a clipping path
                               that the target viewer|printer can handle. */
-    long MaxShadingBitmapSize; /* The maximal number of bytes in
+    int64_t MaxShadingBitmapSize; /* The maximal number of bytes in
                               a bitmap representation of a shading.
                               (Bigger shadings to be downsampled). */
-    long MaxInlineImageSize;
+    int64_t MaxInlineImageSize;
     /* Encryption parameters */
     gs_param_string OwnerPassword;
     gs_param_string UserPassword;
     uint KeyLength;
-    uint Permissions;
+    int32_t Permissions;
     uint EncryptionR;
     gs_param_string NoEncrypt;
     bool EncryptMetadata;
@@ -624,8 +663,9 @@ struct gx_device_pdf_s {
     pdf_page_dsc_info_t page_dsc_info; /* current page */
     /* Additional graphics state */
     bool fill_overprint, stroke_overprint;
+    int rendering_intent;
+    bool user_icc;
     bool remap_fill_color, remap_stroke_color;
-    int overprint_mode;
     gs_id halftone_id;
     gs_id transfer_ids[4];
     int transfer_not_identity;        /* bitmask */
@@ -660,25 +700,33 @@ struct gx_device_pdf_s {
      */
     pdf_temp_file_t streams;
     /*
-     * pictures holds graphic objects being accumulated between BP and EP.
-     * The object is moved to streams when the EP is reached: since BP and
-     * EP nest, we delete the object from the pictures file at that time.
+     * ObjStm holds a tream of objects being stored in an ObjStm, up to MAX_OBJSTM_OBJECTS,
+     * We keep a record of the offset of each object within the stream in ObjStmOffsets, so
+     * that later on we can write the offset within the stream for each object into the
+     * beginning of the ObjStm. NumObjStmOffsets is used to keep track of how many we have,
+     * and ObjStm_id is the id of the stream which we'll eventually store in the 'asides'
+     * file, containing the ObjStm.
      */
-    pdf_temp_file_t pictures;
+    pdf_temp_file_t ObjStm;
+    int64_t ObjStm_id;
+    gs_offset_t *ObjStmOffsets;
+    int NumObjStmObjects;
+    bool doubleXref;
+
     /* ................ */
-    long next_id;
+    int64_t next_id;
     /* The following 3 objects, and only these, are allocated */
     /* when the file is opened. */
     cos_dict_t *Catalog;
     cos_dict_t *Info;
     cos_dict_t *Pages;
 #define pdf_num_initial_ids 3
-    long outlines_id;
+    int64_t outlines_id;
     int next_page;
     int max_referred_page;
-    long contents_id;
+    int64_t contents_id;
     pdf_context_t context;
-    long contents_length_id;
+    int64_t contents_length_id;
     gs_offset_t contents_pos;
     pdf_procset_t procsets;        /* used on this page */
     pdf_text_data_t *text;
@@ -686,7 +734,7 @@ struct gx_device_pdf_s {
 #define initial_num_pages 50
     pdf_page_t *pages;
     int num_pages;
-    ulong used_mask;                /* for where_used: page level = 1 */
+    uint64_t used_mask;                /* for where_used: page level = 1 */
     pdf_resource_list_t resources[NUM_RESOURCE_TYPES];
     /* cs_Patterns[0] is colored; 1,3,4 are uncolored + Gray,RGB,CMYK */
     pdf_resource_t *cs_Patterns[5];
@@ -702,8 +750,13 @@ struct gx_device_pdf_s {
     pdf_article_t *articles;
     cos_dict_t *Dests;
     cos_dict_t *EmbeddedFiles;
+    cos_array_t *AF;
     byte fileID[16];
     /* Use a single time moment for all UUIDs to minimize an indeterminizm. */
+    /* This needs to be a long because we call the platform code gp_get_realtime
+     * to fill it in, and that takes a long *. Be aware that long is 32-bits on
+     * 32-bit Windows and 64-bits on 64-bit Windows.
+     */
     long uuid_time[2];
     /*
      * global_named_objects holds named objects that are independent of
@@ -739,6 +792,7 @@ struct gx_device_pdf_s {
      * redundant clipping paths when PS document generates such ones.
      */
     gx_path *clip_path;
+
     /*
      * Page labels.
      */
@@ -792,6 +846,19 @@ struct gx_device_pdf_s {
     int PatternDepth;
     gs_matrix AccumulatedPatternMatrix;
 
+    /* Normally the PDF itnerpreter doesn't call execform to pass a From XObject to
+     * pdfwrite, but it *does* create forms (and increment FormDepth) for transparent
+     * Groups, which are handled as Form XObjects. Because of the way that Pattterns work,
+     * the first pattern after a Form uses the Form co-ordinate system, but if the Pattern
+     * itself uses Patterns, then the nested Pattern needs to use the co-ordinate system
+     * of the parent Pattern. Unless, of course, we have another Form!
+     * So essentially we need to know the Pattern depth, since the last form was executed.
+     * If its 0, then we want to apply the current CTM, if its more than that then we want
+     * to undo the pdfwrite scaling (see gdevpdfv.c, pdf_store_pattern1_params() at about line
+     * 239.
+     */
+    int PatternsSinceForm;
+
     /* Accessories */
     cos_dict_t *substream_Resources;     /* Substream resources */
     gs_color_space_index pcm_color_info_index; /* Index of the ProcessColorModel space. */
@@ -807,6 +874,7 @@ struct gx_device_pdf_s {
                         Used only with uncached charprocs. */
     bool PS_accumulator; /* A flag to determine whether a given
                          accumulator is for a PostScript type 3 font or not. */
+    bool Scaled_accumulator; /* We scale teh CTM when accumulating type 3 fonts */
     bool accumulating_a_global_object; /* ps2write only.
                         Accumulating a global object (such as a named Form,
                         so that resources used in it must also be global.
@@ -825,6 +893,7 @@ struct gx_device_pdf_s {
     gs_id     image_mask_id;
     bool      image_mask_is_SMask;
     bool      image_mask_skip; /* A flag for pdf_begin_transparency_mask */
+    bool      smask_construction; /* True when pdfwrite is constructing a soft mask */
     uint      image_with_SMask; /* A flag for pdf_begin_transparency_group. In order to
                                  * deal with nested groups we set/test the bit according
                                  * to the FormDepth
@@ -833,6 +902,8 @@ struct gx_device_pdf_s {
     double    image_mask_scale;
     /* Temporary data for soft mask form. */
     pdf_resource_t *pres_soft_mask_dict;
+    /* used to track duplicate stream definitions in pdfmarks */
+    bool pdfmark_dup_stream;
     /* Temporary data for pdfmark_BP. */
     gs_const_string objname;
     int OPDFRead_procset_length;      /* PS2WRITE only. */
@@ -850,7 +921,9 @@ struct gx_device_pdf_s {
     bool DetectDuplicateImages;
     bool AllowIncrementalCFF;
     bool WantsToUnicode;
+    bool PdfmarkCapable;
     bool WantsPageLabels;
+    bool WantsOptionalContent;
     bool AllowPSRepeatFunctions;
     bool IsDistiller;
     bool PreserveSMask;
@@ -893,6 +966,24 @@ struct gx_device_pdf_s {
                                      * doing JPEG pass through we write the JPEG data here, and don't write
                                      * anything in the image processing routines.
                                      */
+    float UserUnit;
+    pdf_OCR_usage UseOCR;           /* Never, AsNeeded or Always */
+    gs_text_enum_t* OCRSaved;       /* Saved state of the text enumerator before rendering glyph bitmaps for later OCR */
+    pdf_OCR_stage OCRStage;         /* Used to control a (sort of) state machine when using OCR to get a Unicode value for a glyph */
+    int *OCRUnicode;                /* Used to pass back the Unicode value from the OCR engine to the text processing */
+    gs_char OCR_char_code;          /* Passes the current character code from text processing to the image processing code when rendering glyph bitmaps for OCR */
+    gs_glyph OCR_glyph;             /* Passes the current glyph code from text processing to the image processing code when rendering glyph bitmaps for OCR */
+    ocr_glyph_t *ocr_glyphs;        /* Records bitmaps and other data from text processing when doing OCR */
+    gs_gstate **initial_pattern_states;
+    bool OmitInfoDate;              /* If true, do not emit CreationDate and ModDate in the Info dictionary and XMP Metadata (must not be true for PDF/X support) */
+    bool OmitXMP;                   /* If true, do not emit an XMP /Metadata block and do not reference it from the Catalog (must not be true for PDF/A output) */
+    bool OmitID;                    /* If true, do not emit a /ID array in the trailer dicionary (must not be true for encrypted files or PDF 2.0) */
+    bool ModifiesPageSize;          /* If true, the new PDF interpreter will not preserve *Box values (the media size has been modified, they will be incorrect) */
+    bool ModifiesPageOrder;         /* If true, the new PDF interpreter will not preserve Outlines or Dests, because they will refer to the wrong page number */
+    bool WriteXRefStm;              /* If true, (the default) use an XRef stream rather than an xref table */
+    bool WriteObjStms;              /* If true, (the default) store candidate objects in ObjStms rather than plain text in the PDF file. */
+    int64_t PendingOC;              /* An OptionalContent object is pending */
+    bool ToUnicodeForStdEnc;        /* Should we emit ToUnicode CMaps when a simple font has only standard glyph names. Defaults to true */
 };
 
 #define is_in_page(pdev)\
@@ -927,7 +1018,7 @@ struct gx_device_pdf_s {
  m(39, gx_device_pdf, EmbeddedFiles);
  m(40, gx_device_pdf, pdf_font_dir);
  m(41, gx_device_pdf, Extension_Metadata);*/
-#define gx_device_pdf_num_ptrs 43
+#define gx_device_pdf_num_ptrs 44
 #define gx_device_pdf_do_param_strings(m)\
     m(0, OwnerPassword) m(1, UserPassword) m(2, NoEncrypt)\
     m(3, DocumentUUID) m(4, InstanceUUID)
@@ -953,6 +1044,7 @@ extern const gx_device_vector_procs pdf_vector_procs;
 dev_proc_fill_rectangle(gdev_pdf_fill_rectangle);
 dev_proc_fill_path(gdev_pdf_fill_path);
 dev_proc_stroke_path(gdev_pdf_stroke_path);
+dev_proc_fill_stroke_path(gdev_pdf_fill_stroke_path);
 dev_proc_fillpage(gdev_pdf_fillpage);
     /* In gdevpdfi.c */
 dev_proc_begin_typed_image(gdev_pdf_begin_typed_image);
@@ -965,7 +1057,7 @@ dev_proc_fill_rectangle_hl_color(gdev_pdf_fill_rectangle_hl_color);
     /* In gdevpdfv.c */
 dev_proc_include_color_space(gdev_pdf_include_color_space);
     /* In gdevpdft.c */
-dev_proc_create_compositor(gdev_pdf_create_compositor);
+dev_proc_composite(gdev_pdf_composite);
 dev_proc_begin_transparency_group(gdev_pdf_begin_transparency_group);
 dev_proc_end_transparency_group(gdev_pdf_end_transparency_group);
 dev_proc_begin_transparency_mask(gdev_pdf_begin_transparency_mask);
@@ -989,6 +1081,10 @@ void pdf_reset_text(gx_device_pdf *pdev);
 
 /* ------ Document ------ */
 
+/* Utility functions to write args into the file as comments */
+int pdfwrite_fwrite_args_comment(gx_device_pdf *pdev, gp_file *f);
+int pdfwrite_write_args_comment(gx_device_pdf *pdev, stream *s);
+
 /* Write a DSC compliant header to the file */
 int ps2write_dsc_header(gx_device_pdf * pdev, int pages);
 
@@ -998,20 +1094,20 @@ int pdfwrite_pdf_open_document(gx_device_pdf * pdev);
 /* ------ Objects ------ */
 
 /* Allocate an ID for a future object, set its pos=0 so we can tell if it is used */
-long pdf_obj_forward_ref(gx_device_pdf * pdev);
+int64_t pdf_obj_forward_ref(gx_device_pdf * pdev);
 
 /* Allocate an ID for a future object. */
-long pdf_obj_ref(gx_device_pdf * pdev);
+int64_t pdf_obj_ref(gx_device_pdf * pdev);
 
 /* Remove an object from the xref table (mark as unused) */
-long pdf_obj_mark_unused(gx_device_pdf *pdev, long id);
+int64_t pdf_obj_mark_unused(gx_device_pdf *pdev, int64_t id);
 
 /* Read the current position in the output stream. */
 gs_offset_t pdf_stell(gx_device_pdf * pdev);
 
 /* Begin an object, optionally allocating an ID. */
-long pdf_open_obj(gx_device_pdf * pdev, long id, pdf_resource_type_t type);
-long pdf_begin_obj(gx_device_pdf * pdev, pdf_resource_type_t type);
+int64_t pdf_open_obj(gx_device_pdf * pdev, int64_t id, pdf_resource_type_t type);
+int64_t pdf_begin_obj(gx_device_pdf * pdev, pdf_resource_type_t type);
 
 /* End an object. */
 int pdf_end_obj(gx_device_pdf * pdev, pdf_resource_type_t type);
@@ -1031,8 +1127,8 @@ extern const char *const pdf_resource_type_names[];
 extern const gs_memory_struct_type_t *const pdf_resource_type_structs[];
 
 /* Record usage of resoruces by pages */
-int pdf_record_usage(gx_device_pdf *const pdev, long resource_id, int page_num);
-int pdf_record_usage_by_parent(gx_device_pdf *const pdev, long resource_id, long parent);
+int pdf_record_usage(gx_device_pdf *const pdev, int64_t resource_id, int page_num);
+int pdf_record_usage_by_parent(gx_device_pdf *const pdev, int64_t resource_id, int64_t parent);
 
 /*
  * Define the offset that indicates that a file position is in the
@@ -1043,18 +1139,24 @@ int pdf_record_usage_by_parent(gx_device_pdf *const pdev, long resource_id, long
 
 /* Begin an object logically separate from the contents. */
 /* (I.e., an object in the resource file.) */
-long pdf_open_separate(gx_device_pdf * pdev, long id, pdf_resource_type_t type);
-long pdf_begin_separate(gx_device_pdf * pdev, pdf_resource_type_t type);
+int64_t pdf_open_separate(gx_device_pdf * pdev, int64_t id, pdf_resource_type_t type);
+int64_t pdf_begin_separate(gx_device_pdf * pdev, pdf_resource_type_t type);
+
+/* functions used for ObjStm writing */
+int FlushObjStm(gx_device_pdf *pdev);
+int NewObjStm(gx_device_pdf *pdev);
+int64_t pdf_open_separate_noObjStm(gx_device_pdf * pdev, int64_t id, pdf_resource_type_t type);
+int pdf_end_separate_noObjStm(gx_device_pdf * pdev, pdf_resource_type_t type);
 
 /* Reserve object id. */
-void pdf_reserve_object_id(gx_device_pdf * pdev, pdf_resource_t *ppres, long id);
+void pdf_reserve_object_id(gx_device_pdf * pdev, pdf_resource_t *ppres, int64_t id);
 
 /* Begin an aside (resource, annotation, ...). */
 int pdf_alloc_aside(gx_device_pdf * pdev, pdf_resource_t ** plist,
                 const gs_memory_struct_type_t * pst, pdf_resource_t **ppres,
-                long id);
+                int64_t id);
 /* Begin an aside (resource, annotation, ...). */
-int pdf_begin_aside(gx_device_pdf * pdev, pdf_resource_t **plist,
+int64_t pdf_begin_aside(gx_device_pdf * pdev, pdf_resource_t **plist,
                     const gs_memory_struct_type_t * pst,
                     pdf_resource_t **ppres, pdf_resource_type_t type);
 
@@ -1068,7 +1170,7 @@ int pdf_begin_resource_body(gx_device_pdf * pdev, pdf_resource_type_t rtype,
 
 /* Allocate a resource, but don't open the stream. */
 int pdf_alloc_resource(gx_device_pdf * pdev, pdf_resource_type_t rtype,
-                       gs_id rid, pdf_resource_t **ppres, long id);
+                       gs_id rid, pdf_resource_t **ppres, int64_t id);
 
 /* Find same resource. */
 int pdf_find_same_resource(gx_device_pdf * pdev,
@@ -1112,7 +1214,7 @@ int pdf_substitute_resource(gx_device_pdf *pdev, pdf_resource_t **ppres,
             bool write);
 
 /* Get the object id of a resource. */
-long pdf_resource_id(const pdf_resource_t *pres);
+int64_t pdf_resource_id(const pdf_resource_t *pres);
 
 /* End a separate object. */
 int pdf_end_separate(gx_device_pdf * pdev, pdf_resource_type_t type);
@@ -1147,13 +1249,18 @@ int pdf_free_resource_objects(gx_device_pdf *pdev, pdf_resource_type_t rtype);
 int pdf_store_page_resources(gx_device_pdf *pdev, pdf_page_t *page, bool clear_usage);
 
 /* Copy data from a temporary file to a stream. */
-int pdf_copy_data(stream *s, FILE *file, gs_offset_t count, stream_arcfour_state *ss);
-int pdf_copy_data_safe(stream *s, FILE *file, gs_offset_t position, long count);
+int pdf_copy_data(stream *s, gp_file *file, gs_offset_t count, stream_arcfour_state *ss);
+int pdf_copy_data_safe(stream *s, gp_file *file, gs_offset_t position, int64_t count);
 
 /* Add the encryption filter. */
 int pdf_begin_encrypt(gx_device_pdf * pdev, stream **s, gs_id object_id);
 /* Remove the encryption filter. */
-void pdf_end_encrypt(gx_device_pdf * pdev);
+static int inline pdf_end_encrypt(gx_device_pdf *pdev)
+{
+    if (pdev->KeyLength)
+        return 1;
+    return 0;
+}
 /* Initialize encryption. */
 int pdf_encrypt_init(const gx_device_pdf * pdev, gs_id object_id, stream_arcfour_state *psarc4);
 
@@ -1161,7 +1268,7 @@ int pdf_encrypt_init(const gx_device_pdf * pdev, gs_id object_id, stream_arcfour
 
 /* Get or assign the ID for a page. */
 /* Returns 0 if the page number is out of range. */
-long pdf_page_id(gx_device_pdf * pdev, int page_num);
+int64_t pdf_page_id(gx_device_pdf * pdev, int page_num);
 
 /* Get the page structure for the current page. */
 pdf_page_t *pdf_current_page(gx_device_pdf *pdev);
@@ -1198,6 +1305,7 @@ typedef struct pdf_lcvd_s {
     gx_device_memory *mask;
     gx_device_pdf *pdev;
     dev_t_proc_copy_color((*std_copy_color), gx_device);
+    dev_t_proc_copy_mono((*std_copy_mono), gx_device);
     dev_t_proc_fill_rectangle((*std_fill_rectangle), gx_device);
     dev_t_proc_close_device((*std_close_device), gx_device);
     dev_t_proc_get_clipping_box((*std_get_clipping_box), gx_device);
@@ -1205,6 +1313,7 @@ typedef struct pdf_lcvd_s {
     bool mask_is_empty;
     bool path_is_empty;
     bool mask_is_clean;
+    bool filled_trap;
     bool write_matrix;
     bool has_background;
     gs_matrix m;
@@ -1219,7 +1328,7 @@ typedef struct pdf_lcvd_s {
 
 int pdf_setup_masked_image_converter(gx_device_pdf *pdev, gs_memory_t *mem, const gs_matrix *m, pdf_lcvd_t **pcvd,
                                  bool need_mask, int x, int y, int w, int h, bool write_on_close);
-int pdf_dump_converted_image(gx_device_pdf *pdev, pdf_lcvd_t *cvd);
+int pdf_dump_converted_image(gx_device_pdf *pdev, pdf_lcvd_t *cvd, int for_pattern);
 void pdf_remove_masked_image_converter(gx_device_pdf *pdev, pdf_lcvd_t *cvd, bool need_mask);
 
 /* ------ Miscellaneous output ------ */
@@ -1278,7 +1387,7 @@ typedef struct pdf_data_writer_s {
     gs_offset_t length_pos;
     pdf_resource_t *pres;
     gx_device_pdf *pdev; /* temporary for backward compatibility of pdf_end_data prototype. */
-    long length_id;
+    int64_t length_id;
     bool encrypted;
 } pdf_data_writer_t;
 /*
@@ -1303,7 +1412,7 @@ int pdf_end_data(pdf_data_writer_t *pdw);
 /* ------ Functions ------ */
 
 /* Define the maximum size of a Function reference. */
-#define MAX_REF_CHARS ((sizeof(long) * 8 + 2) / 3)
+#define MAX_REF_CHARS ((sizeof(int64_t) * 8 + 2) / 3)
 
 /*
  * Create a Function object with or without range scaling.  Scaling means
@@ -1312,10 +1421,6 @@ int pdf_end_data(pdf_data_writer_t *pdw);
  * ranges[i].rmin) / (ranges[i].rmax - ranges[i].rmin).  Note that this is
  * the inverse of the scaling convention for Functions per se.
  */
-#ifndef gs_function_DEFINED
-typedef struct gs_function_s gs_function_t;
-#  define gs_function_DEFINED
-#endif
 int pdf_function(gx_device_pdf *pdev, const gs_function_t *pfn,
                  cos_value_t *pvalue);
 int pdf_function_scaled(gx_device_pdf *pdev, const gs_function_t *pfn,
@@ -1323,7 +1428,7 @@ int pdf_function_scaled(gx_device_pdf *pdev, const gs_function_t *pfn,
 
 /* Write a Function object, returning its object ID. */
 int pdf_write_function(gx_device_pdf *pdev, const gs_function_t *pfn,
-                       long *pid);
+                       int64_t *pid);
 
 /* If a stitching function references an array of other functions, we need
  * to 'unreference' those before freeing the function. otherwise we end up
@@ -1381,9 +1486,9 @@ int pdf_find_named(gx_device_pdf * pdev, const gs_param_string * pname,
  * means just create the object, do not name it.
  */
 int pdf_create_named(gx_device_pdf *pdev, const gs_param_string *pname,
-                     cos_type_t cotype, cos_object_t **ppco, long id);
+                     cos_type_t cotype, cos_object_t **ppco, int64_t id);
 int pdf_create_named_dict(gx_device_pdf *pdev, const gs_param_string *pname,
-                          cos_dict_t **ppcd, long id);
+                          cos_dict_t **ppcd, int64_t id);
 
 /*
  * Look up a named object as for pdf_find_named.  If the object does not
@@ -1513,4 +1618,14 @@ int pdf_from_string_to_text(gx_device_pdf *pdev);
 void pdf_close_text_contents(gx_device_pdf *pdev);
 
 int gdev_pdf_get_param(gx_device *dev, char *Param, void *list);
+
+int pdf_open_temp_file(gx_device_pdf *pdev, pdf_temp_file_t *ptf);
+int pdf_open_temp_stream(gx_device_pdf *pdev, pdf_temp_file_t *ptf);
+int pdf_close_temp_file(gx_device_pdf *pdev, pdf_temp_file_t *ptf, int code);
+
+/* exported by gdevpdfe.c */
+
+int pdf_xmp_write_translated(gx_device_pdf* pdev, stream* s, const byte* data, int data_length,
+    void(*write)(stream* s, const byte* data, int data_length));
+
 #endif /* gdevpdfx_INCLUDED */
